# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from fbprophet import Prophet
import pandas as pd
from datetime import date, timedelta


# Get easter dates
def getEasters(year):
    a = year % 19
    b = year // 100
    c = year % 100
    d = (19 * a + b - b // 4 - ((b - (b + 8) // 25 + 1) // 3) + 15) % 30
    e = (32 + 2 * (b % 4) + 2 * (c // 4) - d - (c % 4)) % 7
    f = d + e - 7 * ((a + 11 * d + 22 * e) // 451) + 114
    month = f // 31
    day = f % 31 + 1
    easter = date(year, month, day)
    return [("easter{}".format(x), easter + timedelta(days=x)) for x in range(-3, 2)]


# Get holidays dataframe in prophet's format
def GetHolidays(years):
    easters = pd.DataFrame({
        'ds': [e[1] for i in years for e in getEasters(i)],
        'holiday': [e[0] for i in years for e in getEasters(i)],
        'lower_window': 0,
        'upper_window': 0,
    })
    return easters


def setupModels(years):
    models = {}
    models["desktop_global"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.7,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.008,
        seasonality_prior_scale=0.20,
        holidays=GetHolidays(years)
    )
    models["nondesktop_global"] = Prophet()
    models["fxa_global"] = Prophet()
    models["desktop_tier1"] = Prophet()
    models["nondesktop_tier1"] = Prophet()
    models["fxa_tier1"] = Prophet()
    return models



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd

import dscontrib.flawrence.bayesian_stats.binary as mabsb


def compare_branches(df, col_label, ref_branch_label='control', thresholds=None):
    """Return the survival functions and relative uplifts thereupon.

    This function generates data for a metric's survival function
    (1 - cumulative distribution function) for each branch, and
    calculates the relative uplift compared to the reference branch
    identified by ``ref_branch_label``.

    It converts the non-negative, real-valued per-user metric data in
    ``df[col_label]`` into ``n=len(thresholds)`` different binary
    metrics, and analyzes these ``n`` metrics with the Bayesian binary
    methods.

    The precise values of the thresholds usually don't matter unless
    certain thresholds have been standardized outside the context of
    this experiment.

    The results are related to those obtained by bootstrapping a range
    of quantiles over the data:

    * In the survival plot, we set a value for the metric and calculate
      the fraction of the data that was above this value, with
      uncertainty on the fraction.
    * When bootstrapping quantiles, we set a quantile (a fraction of
      the data) and find the value such that the given fraction of
      data is greater than this value, with uncertainty on the value.

    Reiterating: if we plot the survival function with metric values
    on the x axis and "fractions" on the y axis, then this function
    first chooses some sensible values for x then runs statistics to
    compute values for y, with uncertainty. If we were bootstrapping
    quantiles, then we would choose some sensible values for y then
    run statistics to compute values for x, with uncertainty.

    Args:
        df: a pandas DataFrame of queried experiment data in the
            standard format. Target metric should be non-negative.
        col_label (str): Label for the df column contaning the metric
            to be analyzed.
        ref_branch_label (str, optional): String in ``df['branch']``
            that identifies the the branch with respect to which we
            want to calculate uplifts - usually the control branch.
        thresholds (list/ndarray, optional): Thresholds that define the
            metric's quantization; ``df[col_label]``

    Returns a dictionary:

        * 'individual': dictionary mapping branch names to a pandas
          DataFrame containing values from the survival function.
          The DataFrames' indexes are the list of thresholds; the
          columns are summary statistics on the survival function.
        * 'comparative': dictionary mapping branch names to a pandas
          DataFrame of summary statistics for the possible uplifts of the
          conversion rate relative to the reference branch - see docs
          for
          :meth:`mozanalysis.stats.summarize_samples.summarize_joint_samples_batch`.
    """
    branch_list = df.branch.unique()

    if not thresholds:
        thresholds = get_thresholds(df[col_label])

    data = {t: _one_thresh(t, df, col_label, ref_branch_label) for t in thresholds}

    return {
        'individual': {
            b: pd.DataFrame({
                t: d['individual'][b] for t, d in data.items()
            }, columns=thresholds).T
            for b in branch_list
        },
        'comparative': {
            b: pd.DataFrame({
                t: d['comparative'][b] for t, d in data.items()
            }).T for b in set(branch_list) - {ref_branch_label}
        },
    }


def get_thresholds(col, max_num_thresholds=101):
    """Return a set of interesting thresholds for the dataset ``col``

    Assumes that the values are non-negative, with zero as a special case.

    Args:
        col: a Series of individuals' data for a metric
        max_num_thresholds (int): Return at most this many threshold values.

    Returns:
        A list of thresholds. By default these are de-duped percentiles
        of the nonzero data.
    """
    if col.isnull().any():
        raise ValueError("'col' contains null values")

    if col.min() < 0:
        raise ValueError("This function assumes non-negative data")

    # When taking quantiles, treat "0" as a special case so that we
    # still have resolution if 99% of users are 0.
    nonzero_quantiles = col[col > 0].quantile(
        np.linspace(0, 1, max_num_thresholds),
        # 'nearest' is not what we want, but is the least-bad option.
        # Can't use the default 'linear' because we want to call 'unique()'
        # to avoid duplicating work.
        # Can't use 'lower', 'higher', or 'midpoint' due to rounding issues
        # that lead to dumb choices. That leaves us with 'nearest'
        interpolation='nearest'
    )
    return sorted(
        [np.float64(0)] + list(nonzero_quantiles.unique())
    )[:-1]  # The thresholds get used as `>` not `>=`, so exclude the max value


def _one_thresh(threshold, df, col_label, ref_branch_label):
    """Run stats on the fraction of clients above ``threshold``."""
    if df[col_label].isnull().any():
        raise ValueError("'df' contains null values for '{}'".format(col_label))

    if '_tmp_threshold_val' in df.columns:
        raise ValueError(
            "Either you have an exceedingly poor taste in column names, "
            "or there is a bug in `_one_thresh`."
        )
    try:
        # Sorry for mutating the input inplace. I'll be sure to tidy up.
        df['_tmp_threshold_val'] = df[col_label] > threshold

        return mabsb.compare_branches(
            df, '_tmp_threshold_val', ref_branch_label=ref_branch_label
        )

    finally:
        df.drop('_tmp_threshold_val', axis='columns', inplace=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import scipy.stats as st
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from dscontrib.flawrence.abtest_stats.beta import compare_two_from_summary


def compare_cdfs(df, col_label, control_label='control'):
    """Compute and plot CDFs and uplifts on CDFs for an A/B test metric

    Args:
        df: a pandas DataFrame of experiment data. Each row represents
            data about an individual test subject. One column is named
            'branch' and contains the test subject's branch. The other
            columns contain the test subject's values for each metric.
        col_label: the name of the column containing the metric of interest
        control_label: the name of the control branch
    """
    df = threshold_and_summarize(df, col_label)

    fig = plt.figure(figsize=(6, 10))
    ax = plt.subplot(211)
    plot_cdf(df, ax, col_label)

    ax = plt.subplot(212)
    plot_relative_differences(df, ax, control_label, col_label)
    return fig


def threshold_and_summarize(df, col_label, thresholds=None):
    """Return values on the CDF for df[col_label] for each branch"""
    if not thresholds:
        thresholds = get_thresholds(df[col_label])

    res = pd.DataFrame(
        index=['num_enrollments'] + thresholds,
        columns=sorted(df.branch.unique())
    )

    for branch in res.columns:
        bdat = df[col_label].loc[df['branch'] == branch]
        res.loc['num_enrollments', branch] = len(bdat)

        for t in thresholds:
            res.loc[t, branch] = (bdat > t).sum()

    return res.astype(np.int64)


def get_thresholds(col, max_num_thresholds=101):
    """Return a set of interesting thresholds for the dataset `col`

    Assumes that the values are non-negative, with zero as a special case.

    Args:
        col: a Series of individual data for a metric
        max_num_thresholds (int): Return at most this many threshold values.

    Returns:
        A list of thresholds. By default these are de-duped percentiles
        of the nonzero data.
    """
    # When taking quantiles, treat "0" as a special case so that we
    # still have resolution if 99% of users are 0.
    nonzero_quantiles = col[col > 0].quantile(
        np.linspace(0, 1, max_num_thresholds)
    )
    return sorted(
        [np.float64(0)] + list(nonzero_quantiles.unique())
    )[:-1]  # The thresholds get used as `>` not `>=`, so exclude the max value


def plot_cdf(pt, ax, xlabel):
    for label, c in pt.items():
        betas = st.beta(
            c.drop('num_enrollments') + 1,
            c.loc['num_enrollments'] - c.drop('num_enrollments') + 1,
        )
        line = ax.plot(c.drop('num_enrollments').index, betas.mean(), label=label)[0]
        col = line.get_color()
        ax.fill_between(
            c.drop('num_enrollments').index.astype('float'),
            betas.ppf(0.05),
            betas.ppf(0.95),
            color=col,
            alpha=0.1
        )
    ax.set_ylim(0, 1)
    ax.legend()
    ax.set_xlabel(xlabel if xlabel is not None else pt.index.name)
    ax.set_ylabel('Fraction of users')
    ax.set_title('1 - CDF: Fraction of users with {} > x'.format(
        'val' if xlabel is not None else pt.index.name)
    )


def plot_relative_differences(pt, ax, control_label, xlabel):
    for c in pt.columns.drop(control_label):
        res = pd.DataFrame([
            compare_two_from_summary(
                pt[[c, control_label]].T,
                control_label=control_label,
                num_conversions_label=x
            )
            for x in pt.index.drop('num_enrollments')
        ], index=pt.index.drop('num_enrollments'))
        line = ax.plot(res.index.astype('float'), res['rel_uplift_exp'], label=c)[0]
        col = line.get_color()
        ax.fill_between(
            res.index.astype('float'),
            res['rel_uplift_0.005'],
            res['rel_uplift_0.995'],
            color=col,
            alpha=0.05
        )
        ax.fill_between(
            res.index.astype('float'),
            res['rel_uplift_0.05'],
            res['rel_uplift_0.95'],
            color=col,
            alpha=0.05
        )
    ax.plot(
        [0, 1, pt.index.drop('num_enrollments').max()],
        [0, 0, 0],
        'k--', label='zero'
    )
    ax.set_xlabel(xlabel if xlabel is not None else pt.index.name)
    ax.set_ylabel('Relative uplift')
    ax.set_title('Uplift in # users >x, relative to {}'.format(control_label))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np


# Calculate Mean Absolute Percentage Error of forecast
def calcMape(true, predicted):
    mask = true != 0
    return (np.fabs(true - predicted)/true)[mask].mean() * 100


# Calculate Mean Relative Error of forecast
def calcMre(true, predicted):
    mask = true != 0
    return ((true - predicted)/true)[mask].mean() * 100


# Get most recent date in table
def getLatestDate(bqClient, project, dataset, table, field):
    query = '''
        SELECT
            MAX({field}) as date
        FROM
            `{project}.{dataset}.{table}`
    '''.format(project=project, dataset=dataset, table=table, field=field)
    return bqClient.query(query).to_dataframe()['date'][0]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mozanalysis.utils import all_

# An easy way to filter for particular event types
metric_library = {
    'session_start': {
        'category': 'action',
        'method': 'foreground',
        'object': 'app',
    },
    'session_end': {
        'category': 'action',
        'method': 'background',
        'object': 'app',
    },
    'erase_button': {
        'category': 'action',
        'method': 'click',
        'object': 'erase_button',
    },
    'back_button': {
        'category': 'action',
        'method': 'click',
        'object': 'back_button',
    },
    'notification': {
        'category': 'action',
        'method': 'click',
        'object': 'notification',
    },
    'notification_erase_open': {
        'category': 'action',
        'method': 'click',
        'object': 'notification_action',
        'value': 'erase_open',
    },
    'home_shortcut': {
        'category': 'action',
        'method': 'click',
        'object': 'notification_action',
        'value': 'erase_open',
    },
    'tabs_tray_erase': {
        'category': 'action',
        'method': 'click',
        'object': 'tabs_tray',
        'value': 'erase',
    },
    'recent_apps_remove': {
        'category': 'action',
        'method': 'click',
        'object': 'recent_apps',
        'value': 'erase',
    }
}


def make_select_col(event, metric_key):
    """Return an int Column named `metric_key`, given an event Column.

    In the returned Column, a row is 1 if the event is `metric_key`,
    otherwise it is 0.

    Useful when trying to count the number of occurrences of an event -
    just sum this column.

    Example usage:

        from pyspark.sql import functions as F

        t = spark.table('telemetry_mobile_event_parquet')
        t2 = t.filter(
            t.submission_date_s3 == '20190101'
        ).select(
            F.explode(t.events).alias('event')
        )
        t3 = t2.select(
            make_select_col(t2.event, 'session_start'),
            make_select_col(t2.event, 'session_end')
        )
        t3.agg(F.sum(t3.session_start), F.sum(t3.session_end)).collect()
    """
    return all_(
        event[k] == v for (k, v) in metric_library[metric_key].items()
    ).astype('int').alias(metric_key)


def make_where(event, metric_key):
    """Return a bool Column named `metric_key`, given an event Column.

    In the returned Column, a row is True iff the event is `metric_key`.

    Useful when filtering for an event.

    Example usage:

        t = spark.table('telemetry_mobile_event_parquet')
        t2 = t.filter(
            t.submission_date_s3 == '20190101'
        ).select(
            F.explode(t.events).alias('event')
        )
        t3 = t2.filter(make_where(t2.event, 'session_start'))
    """
    return all_(
        event[k] == v for (k, v) in metric_library[metric_key].items()
    ).alias(metric_key)



#!/usr/bin/env python
# encoding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime

from setuptools import find_packages, setup


setup(
    name="dscontrib",
    version=datetime.datetime.now().strftime("%Y%m%d%H%M%S"),
    author="Rob Hudson",
    author_email="robhudson@mozilla.com",
    description="A Python library for Mozilla Data Science code snippets",
    url="https://github.com/mozilla/dscontrib",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy",
        "pandas",
        "scipy"
    ],
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd


_queries = {
    "desktop": '''
        SELECT
            submission_date as date,
            sum(mau) AS global_mau,
            SUM(IF(country IN ('US', 'FR', 'DE', 'GB', 'CA'), mau, 0)) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_desktop_exact_mau28_by_dimensions_v1`
        GROUP BY
            date
        ORDER BY
            date
    ''',
    "nondesktop": '''
        SELECT
            submission_date as date,
            sum(mau) AS global_mau,
            SUM(IF(country IN ('US', 'FR', 'DE', 'GB', 'CA'), mau, 0)) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_nondesktop_exact_mau28_by_dimensions_v1`
        GROUP BY
            date
        ORDER BY
            date
    ''',
    "fxa": '''
        SELECT
            submission_date AS date,
            SUM(mau) AS global_mau,
            SUM(seen_in_tier1_country_mau) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_accounts_exact_mau28_by_dimensions_v1`
        GROUP BY
            submission_date
        ORDER BY
            submission_date
    '''
}


def getKpiData(bqClient):
    data = {}
    for q in _queries:
        rawData = bqClient.query(_queries[q]).to_dataframe()
        data['{}_global'.format(q)] = rawData[
            ["date", "global_mau"]
        ].rename(
            index=str, columns={"date": "ds", "global_mau": "y"}
        )
        data['{}_tier1'.format(q)] = rawData[
            ["date", "tier1_mau"]
        ].rename(
            index=str, columns={"date": "ds", "tier1_mau": "y"}
        )
    for k in data:
        data[k]['ds'] = pd.to_datetime(data[k]['ds']).dt.date
    return data



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np
from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from datetime import timedelta

from dscontrib.jmccrosky.forecast.models import setupModels


# Delete output table if necessary and create empty table with appropriate schema
def resetOuputTable(bigquery_client, project, dataset, table_name):
    dataset_ref = bigquery_client.dataset(dataset)
    table_ref = dataset_ref.table(table_name)
    try:
        bigquery_client.delete_table(table_ref)
    except NotFound:
        pass
    schema = [
        bigquery.SchemaField('asofdate', 'DATE', mode='REQUIRED'),
        bigquery.SchemaField('datasource', 'STRING', mode='REQUIRED'),
        bigquery.SchemaField('date', 'DATE', mode='REQUIRED'),
        bigquery.SchemaField('type', 'STRING', mode='REQUIRED'),
        bigquery.SchemaField('mau', 'FLOAT', mode='REQUIRED'),
        bigquery.SchemaField('low90', 'FLOAT', mode='REQUIRED'),
        bigquery.SchemaField('high90', 'FLOAT', mode='REQUIRED'),
    ] + [
        bigquery.SchemaField('p{}'.format(q), 'FLOAT', mode='REQUIRED')
        for q in range(10, 100, 10)
    ]
    table = bigquery.Table(table_ref, schema=schema)
    table = bigquery_client.create_table(table)
    return table


def writeForecasts(bigquery_client, table, model_date, forecast_end, data):
    minYear = np.min([data[k].ds.min() for k in data]).year
    maxYear = forecast_end.year
    years = range(minYear, maxYear + 1)
    models = setupModels(years)
    forecast_start = model_date + timedelta(days=1)
    forecast_period = pd.DataFrame({'ds': pd.date_range(forecast_start, forecast_end)})
    forecast_period['ds']

    for m in data:
        models[m].fit(data[m].query("ds <= @model_date"))
        forecastSamples = models[m].sample_posterior_predictive(
            models[m].setup_dataframe(forecast_period)
        )
        output_data = {
            "asofdate": model_date,
            "datasource": m,
            "date": forecast_period['ds'].dt.date,
            "type": "forecast",
            "mau": np.mean(forecastSamples['yhat'], axis=1),
            "low90": np.nanpercentile(forecastSamples['yhat'], 5, axis=1),
            "high90": np.nanpercentile(forecastSamples['yhat'], 95, axis=1),
        }
        output_data.update({
            "p{}".format(q): np.nanpercentile(forecastSamples['yhat'], q, axis=1)
            for q in range(10, 100, 10)
        })
        errors = bigquery_client.insert_rows(
            table,
            list(pd.DataFrame(output_data).itertuples(index=False, name=None))
        )
        assert errors == []
